@@ -14,26 +14,27 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
+ * Boston, MA  02110-1301  USA.
  */
 
 #include "config.h"
 #include "actions.h"
 #include "exif-i18n.h"
 #include "libjpeg/jpeg-data.h"
+#include "utils.h"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 
 #include <libexif/exif-ifd.h>
 
-#define ENTRY_FOUND     "   *   "
-#define ENTRY_NOT_FOUND "   -   "
-
 #define CN(s) ((s) ? (s) : "(NULL)")
 
+#define TAG_VALUE_BUF 1024
+
 static void
 convert_arg_to_entry (const char *set_value, ExifEntry *e, ExifByteOrder o, ExifLog *log)
 {
@@ -136,6 +137,8 @@ convert_arg_to_entry (const char *set_value, ExifEntry *e, ExifByteOrder o, Exif
 			break;
 		case EXIF_FORMAT_BYTE:
 		case EXIF_FORMAT_SBYTE:
+			e->data[s * i] = atoi ((char *) buf);
+			break;
 		case EXIF_FORMAT_FLOAT:
 		case EXIF_FORMAT_DOUBLE:
 		case EXIF_FORMAT_UNDEFINED:
@@ -190,7 +193,7 @@ show_entry (ExifEntry *entry, unsigned int machine_readable)
 	ExifIfd ifd = exif_entry_get_ifd (entry);
 
 	if (machine_readable) {
-		char b[1024];
+		char b[TAG_VALUE_BUF];
 
 		fprintf (stdout, "%s\n", C(exif_entry_get_value (entry, b, sizeof (b))));
 		return;
@@ -198,7 +201,7 @@ show_entry (ExifEntry *entry, unsigned int machine_readable)
 
 	/*
 	 * The C() macro can point to a static buffer so these printfs
-	 * must be done separately.
+	 * cannot be combined.
 	 */
 	printf (_("EXIF entry '%s' "),
 		C(exif_tag_get_title_in_ifd (entry->tag, ifd)));
@@ -211,18 +214,26 @@ show_entry (ExifEntry *entry, unsigned int machine_readable)
 	exif_entry_dump (entry, 0);
 }
 
-void
-action_set_value (ExifData *ed, ExifLog *log, ExifParams p)
+/*! If the entry doesn't exist, create it. */
+ExifEntry *
+action_create_value (ExifData *ed, ExifLog *log, ExifTag tag, ExifIfd ifd)
 {
 	ExifEntry *e;
 
-	/* If the entry doesn't exist, create it. */
-	if (!((e = exif_content_get_entry (ed->ifd[p.ifd], p.tag)))) {
-	    exif_log (log, EXIF_LOG_CODE_DEBUG, "exif", "Adding entry...");
+	if (!((e = exif_content_get_entry (ed->ifd[ifd], tag)))) {
+	    exif_log (log, EXIF_LOG_CODE_DEBUG, "exif", _("Adding entry..."));
 	    e = exif_entry_new ();
-	    exif_content_add_entry (ed->ifd[p.ifd], e);
-	    exif_entry_initialize (e, p.tag);
+	    exif_content_add_entry (ed->ifd[ifd], e);
+	    exif_entry_initialize (e, tag);
 	}
+	return e;
+}
+
+void
+action_set_value (ExifData *ed, ExifLog *log, ExifParams p)
+{
+	/* If the entry doesn't exist, create it. */
+	ExifEntry *e = action_create_value(ed, log, p.tag, p.ifd);
 
 	/* Now set the value and save the data. */
 	convert_arg_to_entry (p.set_value, e, exif_data_get_byte_order (ed), log);
@@ -277,28 +288,20 @@ action_insert_thumb (ExifData *ed, ExifLog *log, ExifParams p)
 
 	/* Insert new thumbnail */
 	f = fopen (p.set_thumb, "rb");
-	if (!f)
-#ifdef __GNUC__
-		exif_log (log, -1, "exif", _("Could not open "
-			"'%s' (%m)!"), p.set_thumb);
-#else
+	if (!f) {
 		exif_log (log, -1, "exif", _("Could not open "
 			"'%s' (%s)!"), p.set_thumb, strerror (errno));
-#endif
-	fseek (f, 0, SEEK_END);
-	ed->size = ftell (f);
-	ed->data = malloc (sizeof (char) * ed->size);
-	if (ed->size && !ed->data) EXIF_LOG_NO_MEMORY (log, "exif", ed->size);
-	fseek (f, 0, SEEK_SET);
-	if (fread (ed->data, sizeof (char), ed->size, f) != ed->size)
-#ifdef __GNUC__
-		exif_log (log, -1, "exif", _("Could not read "
-			"'%s' (%m)."), p.set_thumb);
-#else
-		exif_log (log, -1, "exif", _("Could not read "
-			"'%s' (%s)."), p.set_thumb, strerror (errno));
-#endif
-	fclose (f);
+	} else {
+		fseek (f, 0, SEEK_END);
+		ed->size = ftell (f);
+		ed->data = malloc (sizeof (char) * ed->size);
+		if (ed->size && !ed->data) EXIF_LOG_NO_MEMORY (log, "exif", ed->size);
+		fseek (f, 0, SEEK_SET);
+		if (fread (ed->data, sizeof (char), ed->size, f) != ed->size)
+			exif_log (log, -1, "exif", _("Could not read "
+				"'%s' (%s)."), p.set_thumb, strerror (errno));
+		fclose (f);
+	}
 }
 
 void
@@ -346,17 +349,14 @@ action_save_thumb (ExifData *ed, ExifLog *log, ExifParams p, const char *fout)
 	/* Save the thumbnail */
 	f = fopen (fout, "wb");
 	if (!f)
-#ifdef __GNUC__
-	exif_log (log, -1, "exif", _("Could not open '%s' for "
-		"writing (%m)!"), fout);
-#else
-	exif_log (log, -1, "exif", _("Could not open '%s' for "
-		"writing (%s)!"), fout, strerror (errno));
-#endif
-	fwrite (ed->data, 1, ed->size, f);
-	fclose (f);
-	fprintf (stdout, _("Wrote file '%s'."), fout);
-	fprintf (stdout, "\n");
+		exif_log (log, -1, "exif", _("Could not open '%s' for "
+			"writing (%s)!"), fout, strerror (errno));
+	else {
+		fwrite (ed->data, 1, ed->size, f);
+		fclose (f);
+		fprintf (stdout, _("Wrote file '%s'."), fout);
+		fprintf (stdout, "\n");
+	}
 }
 
 void
@@ -364,21 +364,44 @@ action_tag_table (ExifData *ed, ExifParams p)
 {
 	unsigned int tag;
 	const char *name;
-	char txt[1024];
-	unsigned int i;
+	char txt[TAG_VALUE_BUF];
+	ExifIfd i;
+	size_t fieldwidth, width, bytes;
+
+#define ENTRY_FOUND     "   *   "
+#define ENTRY_NOT_FOUND "   -   "
 
-	memset (txt, 0, sizeof (txt));
 	snprintf (txt, sizeof (txt) - 1, _("EXIF tags in '%s':"), p.fin);
-	fprintf (stdout, "%-38.38s", txt);
-	for (i = 0; i < EXIF_IFD_COUNT; i++)
-		fprintf (stdout, "%-7.7s", exif_ifd_get_name (i));
+	fieldwidth = width = p.width - 42;
+	bytes = exif_mbstrlen(txt, &width);
+	printf ("%.*s%*s", bytes, txt, fieldwidth-width, "");
+
+	for (i = (ExifIfd)0; i < EXIF_IFD_COUNT; i++) {
+		size_t space;
+		fieldwidth = width = 7;
+		bytes = exif_mbstrlen(exif_ifd_get_name (i), &width);
+		space = fieldwidth-width;
+		printf ("%*s%.*s%*s", space/2, "", bytes, exif_ifd_get_name (i),
+			space - space/2, "");
+	}
 	fputc ('\n', stdout);
+
 	for (tag = 0; tag < 0xffff; tag++) {
-		name = exif_tag_get_title (tag);
+		/*
+		 * Display the name of the first tag of this number found.
+		 * Since there is some overlap (e.g. with GPS tags), this
+		 * name could sometimes be incorrect for the specific tags
+		 * found in this file.
+		 */
+		name = exif_tag_get_title(tag);
 		if (!name)
 			continue;
-		fprintf (stdout, "  0x%04x %-29.29s", tag, C(name));
-		for (i = 0; i < EXIF_IFD_COUNT; i++)
+
+		fieldwidth = width = p.width - 38 - 13;
+		bytes = exif_mbstrlen(C(name), &width);
+		printf ("  0x%04x %.*s%*s",
+			tag, bytes, C(name), fieldwidth-width, "");
+		for (i = (ExifIfd)0; i < EXIF_IFD_COUNT; i++)
 			if (exif_content_get_entry (ed->ifd[i], tag))
 				printf (ENTRY_FOUND);
 			else
@@ -390,21 +413,26 @@ action_tag_table (ExifData *ed, ExifParams p)
 static void
 show_entry_list (ExifEntry *e, void *data)
 {
-	unsigned char *ids = data;
-	char v[128];
+	const ExifParams *p = data;
+	char v[TAG_VALUE_BUF];
 	ExifIfd ifd = exif_entry_get_ifd (e);
+	const char *str;
+	size_t fieldwidth, width, bytes;
 
-	if (*ids)
-		fprintf (stdout, "0x%04x", e->tag);
-	else
-		fprintf (stdout, "%-20.20s", C(exif_tag_get_title_in_ifd (e->tag, ifd)));
+	if (p->use_ids)
+		printf("0x%04x", e->tag);
+	else {
+		str = C(exif_tag_get_title_in_ifd (e->tag, ifd));
+		fieldwidth = width = 20;
+		bytes = exif_mbstrlen(str, &width);
+		printf ("%.*s%*s", bytes, str, fieldwidth-width, "");
+	}
 	printf ("|");
-	if (*ids)
-		fprintf (stdout, "%-72.72s",
-			 C(exif_entry_get_value (e, v, 73)));
-	else
-		fprintf (stdout, "%-58.58s",
-			 C(exif_entry_get_value (e, v, 59)));
+
+	fieldwidth = width = p->use_ids ? p->width-8 : p->width-22;
+	str = C(exif_entry_get_value (e, v, sizeof(v)));
+	bytes = exif_mbstrlen(str, &width);
+	printf("%.*s", bytes, str);
 	fputc ('\n', stdout);
 }
 
@@ -415,29 +443,31 @@ show_ifd (ExifContent *content, void *data)
 }
 
 static void
-print_hline (unsigned char ids)
+print_hline (unsigned char ids, unsigned int screenwidth)
 {
         unsigned int i, width;
 
-        width = (ids ? 6 : 20); 
-        for (i = 0; i < width; i++) fputc ('-', stdout);
+        width = ids ? 6 : 20; 
+        for (i = 0; i < width; i++)
+		fputc ('-', stdout);
         fputc ('+', stdout);
-        for (i = 0; i < 78 - width; i++) fputc ('-', stdout);
+        for (i = 0; i < screenwidth - 2 - width; i++)
+		fputc ('-', stdout);
 	fputc ('\n', stdout);
 }
 
 void
 action_mnote_list (ExifData *ed, ExifParams p)
 {
-	unsigned int i, bs = 1024, c, id;
-	char b[1024];
-	char b1[1024], b2[1024];
+	char b[TAG_VALUE_BUF], b1[TAG_VALUE_BUF], b2[TAG_VALUE_BUF];
+	unsigned int i, c, id;
 	ExifMnoteData *n;
 	const char *s;
+	size_t fieldwidth, width, bytes;
 
 	n = exif_data_get_mnote_data (ed);
 	if (!n) {
-		printf (_("Unknown MakerNote format.\n"));
+		printf (_("Unknown format or nonexistent MakerNote.\n"));
 		return;
 	}
 
@@ -457,22 +487,23 @@ action_mnote_list (ExifData *ed, ExifParams p)
 			sprintf(b1,"0x%04x",id);
 		} else {
 			s = C (exif_mnote_data_get_title (n, i));
-			strncpy (b1, s && *s ? s : _("Unknown tag"), bs);
+			strncpy (b1, s && *s ? s : _("Unknown tag"), TAG_VALUE_BUF);
 			b1[sizeof(b1)-1] = 0;
 		}
-		s = C (exif_mnote_data_get_value (n, i, b, bs));
-		strncpy (b2, s ? s : _("Unknown value"), bs);
-		b2[sizeof(b2)-1] = 0;
-		/* printf ("%s|%s\n", b1, b2); */
-        	if (p.use_ids)
-                	fprintf (stdout, "%-6.6s", b1);
-        	else
-                	fprintf (stdout, "%-20.20s", b1);
+		fieldwidth = width = p.use_ids ? 6 : 20;
+		bytes = exif_mbstrlen(b1, &width);
+                printf ("%.*s%*s", bytes, b1, fieldwidth-width, "");
 		fputc ('|', stdout);
+
+		s = C (exif_mnote_data_get_value (n, i, b, TAG_VALUE_BUF));
+		strncpy (b2, s ? s : _("Unknown value"), TAG_VALUE_BUF);
+		b2[sizeof(b2)-1] = 0;
         	if (p.use_ids) {
 			fputs (b2, stdout);
         	} else {
-                	fprintf (stdout, "%-58.58s", b2);
+			fieldwidth = width = p.width-22;
+			bytes = exif_mbstrlen(b2, &width);
+			printf ("%.*s", bytes, b2);
 		}
         	fputc ('\n', stdout);
 	}
@@ -482,32 +513,42 @@ void
 action_tag_list (ExifData *ed, ExifParams p)
 {
 	ExifByteOrder order;
+	const char *s;
+	size_t fieldwidth, width, bytes;
 
 	if (!ed)
 		return;
 
 	order = exif_data_get_byte_order (ed);
-	fprintf (stdout, _("EXIF tags in '%s' ('%s' byte order):"), p.fin,
+	printf (_("EXIF tags in '%s' ('%s' byte order):"), p.fin,
 		exif_byte_order_get_name (order));
 	fputc ('\n', stdout);
-	print_hline (p.use_ids);
-        if (p.use_ids) {
-                fprintf (stdout, "%-6.6s", _("Tag"));
-        } else {
-                fprintf (stdout, "%-20.20s", _("Tag"));
-        }
+	print_hline (p.use_ids, p.width);
+
+	fieldwidth = width = p.use_ids ? 6 : 20;
+	s = _("Tag");
+	bytes = exif_mbstrlen(s, &width);
+	printf ("%.*s%*s", bytes, s, fieldwidth-width, "");
 	fputc ('|', stdout);
-        if (p.use_ids)
-		fprintf (stdout, "%-72.72s", _("Value"));
-        else
-                fprintf (stdout, "%-58.58s", _("Value"));
+
+	fieldwidth = width = p.use_ids ? p.width-8 : p.width-22;
+	s = _("Value");
+	bytes = exif_mbstrlen(s, &width);
+	printf ("%.*s", bytes, s);
         fputc ('\n', stdout);
-        print_hline (p.use_ids);
-	exif_data_foreach_content (ed, show_ifd, &p.use_ids);
-        print_hline (p.use_ids);
+        print_hline (p.use_ids, p.width);
+
+	if (p.ifd < EXIF_IFD_COUNT)
+		/* Show only a single IFD */
+		show_ifd(ed->ifd[p.ifd], &p);
+	else
+		/* Show contents of all IFDs */
+		exif_data_foreach_content (ed, show_ifd, &p);
+
+        print_hline (p.use_ids, p.width);
         if (ed->size) {
-                fprintf (stdout, _("EXIF data contains a thumbnail "
-				   "(%i bytes)."), ed->size);
+                printf (_("EXIF data contains a thumbnail "
+			  "(%i bytes)."), ed->size);
                 fputc ('\n', stdout);
         }
 }
@@ -516,7 +557,7 @@ static void
 show_entry_machine (ExifEntry *e, void *data)
 {
 	unsigned char *ids = data;
-	char v[1024];
+	char v[TAG_VALUE_BUF];
 	ExifIfd ifd = exif_entry_get_ifd (e);
 
 	if (*ids) {
@@ -540,29 +581,45 @@ action_tag_list_machine (ExifData *ed, ExifParams p)
 {
 	if (!ed) return;
 
-	exif_data_foreach_content (ed, show_ifd_machine, &p.use_ids);
+	if (p.ifd < EXIF_IFD_COUNT)
+		/* Show only a single IFD */
+		show_ifd_machine(ed->ifd[p.ifd], &p.use_ids);
+	else
+		/* Show contents of all IFDs */
+		exif_data_foreach_content (ed, show_ifd_machine, &p.use_ids);
+
 	if (ed->size)
 		fprintf (stdout, _("ThumbnailSize\t%i\n"), ed->size);
 }
 
+/*!
+ * Replace characters which are invalid in an XML tag with safe characters.
+ */
+static inline void
+remove_bad_chars(char *s)
+{
+	while (*s) {
+		if ((*s == '(') || (*s == ')') || (*s == ' '))
+			*s = '_';
+		++s;
+	}
+}
+
 static void
 show_entry_xml (ExifEntry *e, void *data)
 {
 	unsigned char *ids = data;
-	char v[1024], t[1024];
+	char v[TAG_VALUE_BUF], t[TAG_VALUE_BUF];
 
 	if (*ids) {
 		fprintf (stdout, "<0x%04x>", e->tag);
 		fprintf (stdout, "%s", exif_entry_get_value (e, v, sizeof (v)));
 		fprintf (stdout, "</0x%04x>", e->tag);
 	} else {
-		int x;
-		strncpy (t, exif_tag_get_title (e->tag), sizeof (t));
+		strncpy (t, exif_tag_get_title_in_ifd(e->tag, exif_entry_get_ifd(e)), sizeof (t));
 
     /* Remove invalid characters from tag eg. (, ), space */
-		for (x = 0; x < strlen (t); x++)
-			if ((t[x] == '(') || (t[x] == ')') || (t[x] == ' '))
-				t[x] = '_';
+		remove_bad_chars(t);
 
 		fprintf (stdout, "\t<%s>", t);
 		fprintf (stdout, "%s", exif_entry_get_value (e, v, sizeof (v)));
@@ -582,6 +639,11 @@ action_tag_list_xml (ExifData *ed, ExifParams p)
 	if (!ed) return;
 
 	fprintf(stdout, "<exif>\n");
-	exif_data_foreach_content (ed, show_xml, &p.use_ids);
+	if (p.ifd < EXIF_IFD_COUNT)
+		/* Show only a single IFD */
+		show_xml(ed->ifd[p.ifd], &p.use_ids);
+	else
+		/* Show contents of all IFDs */
+		exif_data_foreach_content (ed, show_xml, &p.use_ids);
 	fprintf(stdout, "</exif>\n");
 }
